/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

/* Description: This program sends broadcast messages round in a ring between
 * neighbor groups in two steps (the ring is always composed from some
 * groups of processes and its structure is described in the given mapping ﬁle):
 * -In the first step the root of every even group (local group with even group rank) sends
 *  broadcast to all elements of the right neighbour group.
 * -In the second step the root of every odd group (local group with odd group rank) sends
 *  broadcast to all elements of the right neighbour group.
 *
 * To compile this file use the Makefile located in the directory "examples"
 * and type the command "make broadcastExample"
 *
 * This program requires the following:
 *
 *  - a globus user credentials, you can type:
 * 		grid-proxy-init -verify -debug
 *
 * 	- a machine file with the name "machines" in the same directory
 *    with the following content at least:
 *		"/your_hostname/" 20
 *
 *	- finally the program assumes that the given ring contains even number of
 *    process group (it can be specify in the ring schema).
 *
 * To execute this test using mpirun you can type e.g.:
 * 		mpirun -np 20 broadcastBetweenPairsOfGroups ringWith6Groups.xml
 *
 * 			OR
 *
 * 		../../bin/taagrun ringWith6Groups.xml
 *
 * Remark: In the second case machine file is not required.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mpi.h>
#include <taag.h>

#define MSG_SIZE 160
#define XML_DEFAULT "ringWith6Groups.xml"

void create_message(int rank, int grp, char* s) {
	sprintf(s,"MESSAGE FROM P%d (from G%d)",rank, grp);
}

void process_message(int step, int rank, int grp, char* s) {
	printf("IN STEP %d: P%d (from G%d) RECEIVED: \"%s\".\n", step, rank, grp, s);
}

int main(int argc, char *argv[]) {
	int nrProcs, rc, flag;
	int rank, localRootRank, remoteRootRank;
	int grpRank, leftGrpRank, rightGrpRank;
	MPI_Comm mpiComm1, mpiComm2;
	MPI_Group mpiGrp1, mpiGrp2;

	char buff[MSG_SIZE];

	rc = MPI_Init(&argc,&argv);
	if (rc != MPI_SUCCESS) {
		printf("Error starting MPI program.\n");
		MPI_Abort(MPI_COMM_WORLD, rc);
	}
	MPI_Comm_size(MPI_COMM_WORLD,&nrProcs);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);

	if (argc > 1) {
		rc = TAAG_Init(argv[1]);
	}
	else {
		rc = TAAG_Init(XML_DEFAULT);
	}


	if (rc != TAAG_SUCCESS) {
		fprintf(stderr,"Error (code %d) initializing the TAAG structure on process %d.\n", rc, rank);
		MPI_Finalize();
		return rc;
	} //if

	TAAG_Ring_isRing(&flag);
	if (!flag) {
		fprintf(stderr, "Error (code %d) the given schema is NOT a ring.\n", TAAG_ERR_SCHEMA);
		TAAG_Free();
		MPI_Finalize();
		return TAAG_ERR_SCHEMA;
	}

	TAAG_Group_rank(rank, &grpRank);

	TAAG_Ring_right(grpRank, &rightGrpRank);
	TAAG_Ring_left(grpRank, &leftGrpRank);

	/*********************** First Step ***********************************/
	if (grpRank % 2 == 0) { //even group rank
		TAAG_Group_element(grpRank, 0, &localRootRank);
		TAAG_Group_MPIGroup(1, &localRootRank, 1, &rightGrpRank, &mpiGrp1);

		if (rank == localRootRank) create_message(rank, grpRank, buff);
	} //if
	else { //odd group rank
		TAAG_Group_element(leftGrpRank, 0 , &remoteRootRank);
		TAAG_Group_MPIGroup(1, &remoteRootRank, 1, &grpRank, &mpiGrp1);
	} //else

	//MPI_Comm_create is a collective operation
	MPI_Comm_create(MPI_COMM_WORLD, mpiGrp1, &mpiComm1);

	//sending/receiving broadcast from the root of the even groups to all element of its right neighbour
	if (mpiComm1 != MPI_COMM_NULL) { //not every process is involved in the broadcast
		MPI_Bcast(buff, MSG_SIZE, MPI_CHAR, 0, mpiComm1);
		process_message(1, rank, grpRank, buff);
	}

	/*********************** Second Step **********************************/
	if (grpRank % 2 == 1) { //odd group rank
		TAAG_Group_element(grpRank, 0, &localRootRank);
		TAAG_Group_MPIGroup(1, &localRootRank, 1, &rightGrpRank, &mpiGrp2);

		if (rank == localRootRank) create_message(rank, grpRank, buff);
	} //if
	else { //even group rank
		TAAG_Group_element(leftGrpRank, 0 , &remoteRootRank);
			TAAG_Group_MPIGroup(1, &remoteRootRank, 1, &grpRank, &mpiGrp2);
	} //else

	//MPI_Comm_create is a collective operation
	MPI_Comm_create(MPI_COMM_WORLD, mpiGrp2, &mpiComm2);

	//sending/receiving broadcast from the root of the even groups to all element of its right neighbour
	if (mpiComm2 != MPI_COMM_NULL) { //not every process is involved in the broadcast
		MPI_Bcast(buff, MSG_SIZE, MPI_CHAR, 0, mpiComm2);
		process_message(2, rank, grpRank, buff);
	}

	if (mpiComm1 != MPI_COMM_NULL) MPI_Comm_free(&mpiComm1);
	if (mpiComm2 != MPI_COMM_NULL) MPI_Comm_free(&mpiComm2);
	MPI_Group_free(&mpiGrp1);
	MPI_Group_free(&mpiGrp2);

	TAAG_Free();
	MPI_Finalize();
	return 0;
}
